@@ -1,25 +1,36 @@
 import copy
 import os
+import time
 from pprint import pprint
 
-from octopus_python_client.helper import compare_overwrite, find_item, compare_dicts, load_file, save_file
-from octopus_python_client.send_requests_to_octopus import call_octopus, operation_get, operation_post, operation_put, \
+from octopus_python_client.utilities.helper import compare_overwrite, find_item, compare_dicts, load_file, save_file
+from octopus_python_client.utilities.send_requests_to_octopus import call_octopus, operation_get, operation_post, \
+    operation_put, \
     operation_delete
 
 # constants
 all_underscore = "all_"
+error_message_key = "ErrorMessage"
+error_message_resource_not_found = "The resource you requested was not found."
+comma_sign = ","
 dot_sign = "."
 double_hyphen = "--"
+environments_prefix = "Environments"
+executing_string = "Executing"
 file_configuration = "configuration.json"
 folder_outer_spaces = "outer_spaces"
 folder_configurations = "configurations"
 hyphen_sign = "-"
+runbook_process_prefix = "RunbookProcess"
 slash_all = "/all"
 slash_sign = "/"
 space_map = "space_map"
+success_string = "Success"
+tenants_prefix = "Tenants"
 underscore_sign = "_"
 url_all_pages = "?skip=0&take=2147483647"
 yaml_ext = ".yaml"
+positive_integer_regex = "-[1-9][0-9]*$"
 
 # dict keys
 action_name_key = "ActionName"
@@ -27,7 +38,12 @@ actions_key = 'Actions'
 api_key_key = "api_key"
 canonical_tag_name_key = "CanonicalTagName"
 channel_id_key = "ChannelId"
+cloned_from_tenant_id_key = "ClonedFromTenantId"
+comments_key = "Comments"
 deployment_process_id_key = 'DeploymentProcessId'
+description_key = "Description"
+donor_package_key = "DonorPackage"
+donor_package_step_id_key = "DonorPackageStepId"
 environment_id_key = "EnvironmentId"
 environment_ids_key = "EnvironmentIds"
 feed_id_key = "FeedId"
@@ -38,22 +54,33 @@ items_key = 'Items'
 life_cycle_id_key = "LifecycleId"
 name_key = 'Name'
 new_value_key = "NewValue"
+next_version_increment_key = "NextVersionIncrement"
 octopus_endpoint_key = "octopus_endpoint"
 octopus_name_key = "octopus_name"
 owner_id_key = "OwnerId"
+package_id_key = "PackageId"
 package_reference_name_key = "PackageReferenceName"
+packages_key = "Packages"
 password_key = "password"
 project_group_id_key = "ProjectGroupId"
 project_id_key = "ProjectId"
+published_runbook_snapshot_id_key = "PublishedRunbookSnapshotId"
 release_id_key = "ReleaseId"
 release_notes_key = "ReleaseNotes"
+runbook_id_key = "RunbookId"
+runbook_process_id_key = "RunbookProcessId"
 secret_key_key = "SecretKey"
 selected_packages_key = "SelectedPackages"
+state_key = "State"
 steps_key = 'Steps'
 space_id_key = "SpaceId"
 tags_key = "Tags"
+task_id_key = "TaskId"
+team_id_key = "TeamId"
 tenant_id_key = "TenantId"
+token_key = "Token"
 user_name_key = "user_name"
+user_role_id_key = "UserRoleId"
 value_key = "Value"
 variables_key = "Variables"
 variable_set_id_key = "VariableSetId"
@@ -64,7 +91,9 @@ worker_pool_id_key = "WorkerPoolId"
 # item types
 item_type_accounts = "accounts"
 item_type_action_templates = "actiontemplates"
+item_type_api_keys = "apikeys"
 item_type_artifacts = "artifacts"
+item_type_build_information = "build-information"
 item_type_channels = "channels"
 item_type_certificates = "certificates"
 item_type_configuration = "configuration"
@@ -84,11 +113,16 @@ item_type_machine_policies = "machinepolicies"
 item_type_machine_roles = "machineroles"
 item_type_machines = "machines"
 item_type_packages = "packages"
+item_type_permissions = "permissions"
+item_type_permissions_configuration = "permissions/configuration"
+item_type_permissions_export = "permissions/export"
 item_type_project_groups = "projectgroups"
 item_type_project_triggers = "projecttriggers"
 item_type_projects = "projects"
 item_type_proxies = "proxies"
 item_type_releases = "releases"
+item_type_runbook_processes = "runbookProcesses"
+item_type_runbook_snapshots = "runbookSnapshots"  # clone requires package version, so not working now
 item_type_runbooks = "runbooks"
 item_type_scoped_user_roles = "scopeduserroles"  # the space id should not be null in the response, buggy!
 item_type_spaces = 'spaces'
@@ -101,11 +135,17 @@ item_type_teams = "teams"
 item_type_tenants = "tenants"
 item_type_tenant_variables = "tenantvariables"
 item_type_user_onboarding = "useronboarding"
+item_type_users = "users"
 item_type_variables = "variables"
 item_type_variables_names = "variables/names"
 item_type_worker_pools = "workerpools"
 item_type_workers = "workers"
 
+# some types have the extended types like: /api/users/{id}/permissions
+user_ext_types = [item_type_api_keys, item_type_permissions, item_type_permissions_configuration,
+                  item_type_teams]  # item_type_permissions_export is csv file
+ext_types_map = {item_type_users: user_ext_types}
+
 item_types_with_duplicate_names = \
     {item_type_channels, item_type_tasks, item_type_deployments, item_type_configuration, item_type_spaces}
 
@@ -115,37 +155,41 @@ item_types_without_single_item = \
 # must have for a space to work
 must_have_types = [item_type_environments]
 # these types do not have dependencies on other types
-basic_types = [item_type_action_templates, item_type_certificates,
-               item_type_dashboard_configuration, item_type_feeds, item_type_machine_policies,
-               item_type_machines, item_type_proxies, item_type_subscriptions, item_type_tag_sets, item_type_teams,
-               item_type_library_variable_sets, item_type_worker_pools]
+basic_types = [item_type_action_templates, item_type_certificates, item_type_feeds, item_type_machine_policies,
+               item_type_machines, item_type_proxies, item_type_subscriptions, item_type_tag_sets,
+               item_type_teams, item_type_library_variable_sets, item_type_worker_pools]
 # these types have links/dependencies on other types
 complex_types = [item_type_workers, item_type_life_cycles, item_type_project_groups, item_type_projects,
-                 item_type_tenants, item_type_channels, item_type_project_triggers]
+                 item_type_runbooks, item_type_tenants, item_type_channels, item_type_project_triggers,
+                 item_type_accounts, item_type_build_information]
+# these types have only one item in a space; they are mostly space properties
+space_single_item_types = [item_type_dashboard_configuration]
 # these types are the child type of another type
-child_types = [item_type_deployment_processes]
+child_types = [item_type_deployment_processes, item_type_runbook_processes]
 # these types needs "/all" to get all items for this type
-only_all_types = [item_type_variables, item_type_tenant_variables, item_type_machine_roles]
-# the other types not cloneable
-other_types = [item_type_accounts, item_type_packages, item_type_releases, item_type_interruptions,
-               item_type_user_onboarding, item_type_dashboard, item_type_dashboard_dynamic, item_type_deployments,
-               item_type_variables_names, item_type_runbooks, item_type_artifacts]
+only_all_types_inside_space = [item_type_variables, item_type_tenant_variables, item_type_machine_roles]
+# the other types not cloneable for now
+other_types = [item_type_packages, item_type_releases, item_type_interruptions, item_type_user_onboarding,
+               item_type_dashboard, item_type_dashboard_dynamic, item_type_deployments, item_type_variables_names,
+               item_type_artifacts, item_type_home, item_type_scoped_user_roles, item_type_runbook_snapshots]
 # too many items in them, so ignore for now
 large_types = [item_type_tasks, item_type_events]
 # the types which are cloneable
 normal_cloneable_types = must_have_types + basic_types + complex_types
 # the types live inside space
-item_types_inside_space = normal_cloneable_types + child_types + only_all_types + other_types
+item_types_inside_space = normal_cloneable_types + child_types + only_all_types_inside_space + other_types
 # the types live outside space (Octopus server types)
 item_types_only_ourter_space = \
     ["authentication", "configuration/certificates", "communityactiontemplates", "externalsecuritygroupproviders",
-     "featuresconfiguration", "letsencryptconfiguration", "licenses/licenses-current", "maintenanceconfiguration",
-     "octopusservernodes", "performanceconfiguration", "permissions/all", "scheduler", "serverconfiguration",
-     "serverconfiguration/settings", "serverstatus", "smtpconfiguration", "upgradeconfiguration", "users",
-     "userroles", item_type_configuration, item_type_spaces, item_type_home, item_type_scoped_user_roles]
+     "featuresconfiguration", "letsencryptconfiguration", "licenses/licenses-current",
+     "licenses/licenses-current-status", "maintenanceconfiguration", "octopusservernodes", "performanceconfiguration",
+     "permissions/all", "scheduler", "serverconfiguration", "serverconfiguration/settings", "serverstatus",
+     "smtpconfiguration", "smtpconfiguration/isconfigured", "upgradeconfiguration", item_type_users, "userroles",
+     item_type_configuration, item_type_spaces]
 # the sub item map for a specific type; this is for deleting the unused sub items when the item cannot be deleted
 # e.g. tagsets: Tags is the key to get the list of the sub items; CanonicalTagName is for printing purpose
 item_type_sub_item_map = {item_type_tag_sets: (tags_key, canonical_tag_name_key)}
+item_id_prefix_to_type_dict = {environments_prefix: item_type_environments, tenants_prefix: item_type_tenants}
 
 
 class Config:
@@ -155,34 +199,36 @@ class Config:
         self.api_key = None
         self.user_name = None
         self.password = None
-        self.code_path = os.path.dirname(os.path.abspath(__file__))
-        self.current_path = os.getcwd()
         self.overwrite = False
+        self.current_path = None
         self.get_config()
 
     def get_config(self):
-        print("********** Octopus deploy python client tool **********")
-        print('code_path: ' + self.code_path)
-        print('current working path: ' + self.current_path)
-        config_file = os.path.join(self.code_path, folder_configurations, file_configuration)
+        code_path = os.path.dirname(os.path.abspath(__file__))
+        config_file = os.path.join(code_path, folder_configurations, file_configuration)
         config_dict = load_file(config_file)
         self.octopus_endpoint = config_dict.get(octopus_endpoint_key)
         self.octopus_name = config_dict.get(octopus_name_key)
         self.api_key = config_dict.get(api_key_key)
         self.user_name = config_dict.get(user_name_key)
         self.password = config_dict.get(password_key)
+        self.current_path = os.getcwd()
 
 
 config = Config()
 
 
-def verify_space(space_id_or_name=None):
-    all_spaces = get_one_type(item_type=item_type_spaces)
-    list_spaces = get_list_items_from_all_items(all_items=all_spaces)
-    space = find_item(lst=list_spaces, key=id_key, value=space_id_or_name)
+def get_list_ids_one_type(item_type=None, space_id=None):
+    list_items = get_one_type_to_list(item_type=item_type, space_id=space_id)
+    return [item.get(id_key) for item in list_items]
+
+
+def verify_space(space_id_name=None):
+    list_spaces = get_one_type_to_list(item_type=item_type_spaces)
+    space = find_item(lst=list_spaces, key=id_key, value=space_id_name)
     if space:
         return space.get(id_key)
-    space = find_single_item_from_list_by_name(list_items=list_spaces, item_name=space_id_or_name)
+    space = find_single_item_from_list_by_name(list_items=list_spaces, item_name=space_id_name)
     if space:
         return space.get(id_key)
     return None
@@ -190,8 +236,9 @@ def verify_space(space_id_or_name=None):
 
 # remove the unnecessary modified date/user information from put and post operations
 def pop_last_modified(a_dict=None):
-    a_dict.pop('LastModifiedOn', None)
-    a_dict.pop('LastModifiedBy', None)
+    if isinstance(a_dict, dict):
+        a_dict.pop('LastModifiedOn', None)
+        a_dict.pop('LastModifiedBy', None)
 
 
 def always_overwrite_or_compare_overwrite(local_file=None, data=None, overwrite=False):
@@ -291,36 +338,55 @@ def compare_overwrite_multiple_items(items=None, item_type=None, space_id=None,
 
 # get all items for an item_type by call Octopus API /api/{space_id}/item_type with 'get' operation
 # {space_id} is optional
-def get_one_type(item_type=None, space_id=None):
+def get_one_type_ignore_error(item_type=None, space_id=None):
     if not item_type:
         raise ValueError("item_type must not be empty")
     space_url = space_id + slash_sign if space_id else ""
-    if item_type == item_type_home:
-        return call_octopus(config=config, url_suffix=space_url)
-    if item_type in only_all_types:
-        url_suffix = space_url + item_type + slash_all
-        # TODO bug https://help.octopus.com/t/504-gateway-time-out-on-getting-all-variables/24732
-        try:
+    try:
+        if item_type == item_type_home:
+            return call_octopus(config=config, url_suffix=space_url)
+        if item_type in only_all_types_inside_space:
+            url_suffix = space_url + item_type + slash_all
             return call_octopus(config=config, url_suffix=url_suffix)
-        except ValueError as err:
-            print(err)
-            return {}
-    else:
-        url_suffix = space_url + item_type + url_all_pages
-        return call_octopus(config=config, url_suffix=url_suffix)
+        else:
+            url_suffix = space_url + item_type + url_all_pages
+            return call_octopus(config=config, url_suffix=url_suffix)
+    except ValueError as err:
+        print(err)
+        # TODO bug https://help.octopus.com/t/504-gateway-time-out-on-getting-all-variables/24732
+        print(f"resource exist for {item_type} in {space_id}, but some error prevents from getting the resource")
+        return {}
+
+
+# get extended types like /api/users/{id}/permissions
+def get_ext_types_save(item_type=None, space_id=None, item_ids=None):
+    ext_types = ext_types_map.get(item_type)
+    print(f"Get extended types {ext_types} of {item_type} in Space {space_id}")
+    for ext_type in ext_types:
+        ext_items_dict = {}
+        for item_id in item_ids:
+            address = item_type + slash_sign + item_id + slash_sign + ext_type
+            ext_item = request_octopus_item(payload=space_id, space_id=space_id, address=address)
+            ext_items_dict[item_id] = ext_item
+        ext_file = get_local_single_item_file(item_name=all_underscore + item_type + underscore_sign + ext_type,
+                                              item_type=item_type, space_id=space_id)
+        save_file(file_path_name=ext_file, content=ext_items_dict)
 
 
-# get/delete all items for an item_type by call Octopus API /api/{space_id}/item_type
 # then save the all items into a local file (warning for overwrite)
 def get_one_type_save(item_type=None, space_id=None, overwrite=False):
     if not item_type:
         raise ValueError("item_type must not be empty")
-    all_items = get_one_type(item_type=item_type, space_id=space_id)
+    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
     compare_overwrite_multiple_items(items=all_items, item_type=item_type, space_id=space_id, overwrite=overwrite)
+    if item_type == item_type_users:
+        list_items = get_list_items_from_all_items(all_items=all_items)
+        item_ids = [item.get(id_key) for item in list_items]
+        get_ext_types_save(item_type=item_type_users, space_id=space_id, item_ids=item_ids)
     return all_items
 
 
-# get/delete all items for an item_type by call Octopus API /api/{space_id}/item_type
+# delete all items for an item_type by call Octopus API /api/{space_id}/item_type
 # then save the all items into a local file (warning for overwrite)
 def delete_one_type(item_type=None, space_id=None):
     if not item_type:
@@ -330,7 +396,7 @@ def delete_one_type(item_type=None, space_id=None):
             config.overwrite = True
         elif input(f"Delete NONE items of {item_type} in {space_id} [Y/n]? ") == 'Y':
             return
-    all_items = get_one_type(item_type=item_type, space_id=space_id)
+    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
     if item_type in item_types_without_single_item:
         print(f"{item_type} has no sub-single-item, exit")
         return
@@ -340,17 +406,19 @@ def delete_one_type(item_type=None, space_id=None):
 
 # get all items for all item_type(s) by call Octopus API /api/{space_id}/item_type with 'get' operation
 # item_types can be None, "", or "projects,tenants" etc
-def get_all_item_types_save(item_types_comma_delimited=None, space_id=None):
+def get_types_save(item_types_comma_delimited=None, space_id=None):
     if item_types_comma_delimited:
-        list_item_types = item_types_comma_delimited.split(",")
+        list_item_types = item_types_comma_delimited.split(comma_sign)
     else:
         if space_id:
             list_item_types = item_types_inside_space
         else:
             list_item_types = item_types_inside_space + item_types_only_ourter_space
-    if not config.overwrite:
+    if config.overwrite:
+        print(f"===== You are downloading {list_item_types} from space {space_id}... ===== ")
+    else:
         config.overwrite = \
-            input(f"***** You are downloading {list_item_types} from {space_id}; "
+            input(f"***** You are downloading {list_item_types} from space {space_id}; "
                   f"Some entities may already be downloaded locally; "
                   f"Do you want to overwrite all local existing entities? "
                   f"If no, you will be asked to overwrite or not for each type respectively. [Y/n]: ") == 'Y'
@@ -358,6 +426,26 @@ def get_all_item_types_save(item_types_comma_delimited=None, space_id=None):
         get_one_type_save(item_type=item_type, space_id=space_id)
 
 
+def get_spaces_save(item_types_comma_delimited=None, space_id_or_name_comma_delimited=None):
+    if space_id_or_name_comma_delimited:
+        list_space_ids_or_names = space_id_or_name_comma_delimited.split(comma_sign)
+        list_space_ids = [verify_space(space_id_name=space_id_or_name) for space_id_or_name in
+                          list_space_ids_or_names]
+    else:
+        list_space_ids = get_list_ids_one_type(item_type=item_type_spaces) + [None]
+    list_space_ids_set = set(list_space_ids)
+    if config.overwrite:
+        print(f"===== You are downloading spaces {list_space_ids_set}... =====")
+    else:
+        config.overwrite = \
+            input(f"===== You are downloading spaces {list_space_ids_set}; "
+                  f"Some entities may already be downloaded locally; "
+                  f"Do you want to overwrite all local existing entities? "
+                  f"If no, you will be asked to overwrite or not for each type respectively. [Y/n]: ") == 'Y'
+    for space_id in list_space_ids_set:
+        get_types_save(item_types_comma_delimited=item_types_comma_delimited, space_id=space_id)
+
+
 # get a single item from Octopus server
 # 1. get all items for an item_type by call Octopus API /api/{space_id}/item_type with 'get' operation
 # 2. find the matching item for the item_name
@@ -366,7 +454,7 @@ def get_single_item_by_name(item_type=None, item_name=None, space_id=None):
         raise ValueError("item_type and item_name must not be empty")
     print(f"Getting {item_type} {item_name} from {space_id} "
           f"by getting all items first and then find the matched item by name")
-    all_items = get_one_type(item_type=item_type, space_id=space_id)
+    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
     return find_single_item_from_list_by_name(list_items=all_items.get(items_key, []), item_name=item_name)
 
 
@@ -478,9 +566,9 @@ def get_or_delete_single_item_by_id(item_type=None, item_id=None, action=operati
 
 def get_list_items_from_all_items(all_items=None):
     # the case where the payload has a metadata and a list
-    if isinstance(all_items, dict) and all_items.get(items_key):
+    if isinstance(all_items, dict) and isinstance(all_items.get(items_key), list):
         print(f"the payload is a dict, so get the list of items first by {items_key}")
-        return all_items.get(items_key, [])
+        return all_items.get(items_key)
     elif isinstance(all_items, list):
         print("the payload is a list")
         return all_items
@@ -515,7 +603,8 @@ def put_single_item(item_type=None, payload=None, space_id=None):
     if not item_type or not payload:
         raise ValueError("item_type and playload must not be empty")
     space_url = space_id + slash_sign if space_id else ""
-    url_suffix = space_url + item_type + slash_sign + payload.get(id_key)
+    # some type has no id like http://server/api/Spaces-1/dashboardconfiguration
+    url_suffix = space_url + item_type + (slash_sign + payload.get(id_key) if payload.get(id_key) else "")
     item = call_octopus(operation=operation_put, payload=payload, config=config, url_suffix=url_suffix)
     pop_last_modified(item)
     print(f"{item_type} {id_key} is " + item[id_key])
@@ -802,10 +891,28 @@ def get_list_variables_by_set_name_or_id(set_name=None, set_id=None, space_id=No
     return variables_dict.get(variables_key)
 
 
-def get_one_type_to_list_ignore_error(item_type=None, space_id=None):
-    all_items = None
-    try:
-        all_items = get_one_type(item_type=item_type, space_id=space_id)
-    except ValueError as err:
-        print(err)
+def get_one_type_to_list(item_type=None, space_id=None):
+    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
     return get_list_items_from_all_items(all_items=all_items)
+
+
+def get_task_status(task_id=None, space_id=None):
+    print(f"check the status of task {task_id} in space {space_id}")
+    task = get_or_delete_single_item_by_id(item_type=item_type_tasks, item_id=task_id, space_id=space_id)
+    save_single_item(item_type=item_type_tasks, item=task, space_id=space_id)
+    print(f"the task's status is {task.get(state_key)} and description is: {task.get(description_key)}")
+    return task.get(state_key)
+
+
+def wait_task(task_id=None, space_id=None, time_limit_second=600):
+    print(f"wait for task {task_id} in space {space_id} to complete until time out at {time_limit_second} seconds")
+    counter = 0
+    while get_task_status(task_id=task_id, space_id=space_id) == executing_string:
+        if counter > time_limit_second:
+            print(f"task {task_id} takes longer than {time_limit_second} seconds and times out")
+            return executing_string
+        time.sleep(1)
+        counter += 1
+    status = get_task_status(task_id=task_id, space_id=space_id)
+    print(f"task {task_id} in space {space_id} completes with status {status} at {counter} seconds")
+    return status
