@@ -1,17 +1,22 @@
 import copy
+import re
 from pprint import pprint
 from time import gmtime, strftime
 
 from octopus_python_client.common import name_key, tags_key, id_key, item_type_tag_sets, post_single_item_save, \
     item_type_projects, get_list_items_from_file, item_type_deployment_processes, deployment_process_id_key, config, \
-    get_one_type_save, versioning_strategy_key, item_types_inside_space, put_single_item_save, normal_cloneable_types, \
+    get_one_type_save, versioning_strategy_key, put_single_item_save, normal_cloneable_types, \
     get_list_items_from_all_items, version_key, must_have_types, item_type_library_variable_sets, item_type_variables, \
     variable_set_id_key, get_or_delete_single_item_by_id, item_type_tenant_variables, canonical_tag_name_key, \
     item_type_tags, tenant_id_key, item_type_migration, space_map, get_local_single_item_file, put_single_item, \
     get_single_item_by_name_or_id_save, item_type_tenants, slash_sign, underscore_sign, item_type_feeds, \
-    secret_key_key, new_value_key, hyphen_sign, get_one_type, item_type_channels, project_id_key, item_type_releases, \
-    item_type_artifacts, file_name_key, put_post_tenant_variables_save, get_one_type_to_list_ignore_error
-from octopus_python_client.helper import find_item, save_file, find_intersection_multiple_keys_values
+    secret_key_key, new_value_key, hyphen_sign, item_type_channels, project_id_key, \
+    item_type_releases, item_type_artifacts, file_name_key, put_post_tenant_variables_save, \
+    get_one_type_to_list, donor_package_key, runbook_process_id_key, item_type_runbooks, \
+    donor_package_step_id_key, item_type_accounts, token_key, comma_sign, space_id_key, \
+    published_runbook_snapshot_id_key, item_type_scoped_user_roles, user_role_id_key, team_id_key, \
+    item_type_runbook_processes, runbook_process_prefix, item_id_prefix_to_type_dict, positive_integer_regex
+from octopus_python_client.utilities.helper import find_item, save_file, find_matched_sub_list
 
 
 class Migration:
@@ -29,28 +34,34 @@ class Migration:
         self.__type_post_func_dict = {}
         self.__type_prep_func_dict = {}
         self.__type_src_list_items_dict = {}
+        self.__item_types_set = set()
 
     # search the type in the space and see if the matched item already exists
-    def __find_matched_dst_item_by_src_item(self, item=None, item_type=None):
-        dst_all_items = get_one_type(item_type=item_type, space_id=self.__dst_space_id)
-        dst_list_items = get_list_items_from_all_items(all_items=dst_all_items)
-        item_name = item.get(name_key)
-        # "channels" type is special, the name is not unique: channel "Default" could be in different projects
-        # so we use both channel name and project id to find the match
-        if item_type == item_type_channels:
-            keys_values = {name_key: item_name, project_id_key: item.get(project_id_key)}
+    def __find_matched_dst_item_by_src_item(self, src_item_dst_ids=None, item_type=None):
+        print(f"Look for the matched destination {item_type} {src_item_dst_ids.get(name_key)} in {self.__dst_space_id}")
+        dst_list_items = get_one_type_to_list(item_type=item_type, space_id=self.__dst_space_id)
+        item_name = src_item_dst_ids.get(name_key)
+        # "channels" and "runbooks" are special, the name is not unique across a space;
+        # we must use both name and project id to find the match
+        if item_type == item_type_channels or item_type == item_type_runbooks:
+            match_dict = {name_key: item_name, project_id_key: src_item_dst_ids.get(project_id_key)}
         # type "releases" has no name and is unique by "Version" and "ProjectId"
         elif item_type == item_type_releases:
-            keys_values = {version_key: item.get(version_key), project_id_key: item.get(project_id_key)}
+            match_dict = {version_key: src_item_dst_ids.get(version_key),
+                          project_id_key: src_item_dst_ids.get(project_id_key)}
+        # TODO cloning scopeduserroles is not working and may not be necessary; some space id is null;
+        elif item_type == item_type_scoped_user_roles:
+            match_dict = {user_role_id_key: src_item_dst_ids.get(user_role_id_key),
+                          team_id_key: src_item_dst_ids.get(team_id_key)}
         # TODO type "artifacts" has no name and unique by Filename and ServerTaskId, so it is not cloneable
         elif item_type == item_type_artifacts:
-            return find_item(lst=dst_list_items, key=file_name_key, value=item.get(file_name_key))
+            return find_item(lst=dst_list_items, key=file_name_key, value=src_item_dst_ids.get(file_name_key))
         elif item_name:
-            keys_values = {name_key: item_name}
+            match_dict = {name_key: item_name}
         else:
-            pprint(item)
+            pprint(src_item_dst_ids)
             raise ValueError(f"{item_type} does not have name or other keys for matching!")
-        intersection_list = find_intersection_multiple_keys_values(lst=dst_list_items, keys_values=keys_values)
+        intersection_list = find_matched_sub_list(lst=dst_list_items, match_dict=match_dict)
         if intersection_list and len(intersection_list) > 1:
             pprint(intersection_list)
             raise ValueError(f"For {item_name}, more than one item in {self.__dst_space_id} are found")
@@ -66,7 +77,7 @@ class Migration:
         # the source tag id should be all removed
         for src_tag in src_list_tags:
             src_tag.pop(id_key, None)
-        dst_tag_set = self.__find_matched_dst_item_by_src_item(item=src_item, item_type=item_type_tag_sets)
+        dst_tag_set = self.__find_matched_dst_item_by_src_item(src_item_dst_ids=src_item, item_type=item_type_tag_sets)
         if dst_tag_set:
             dst_list_tags = dst_tag_set.get(tags_key, [])
             for src_tag in src_list_tags:
@@ -78,9 +89,11 @@ class Migration:
     # child items will be created automatically when parent item is created
     def __prepare_project(self, src_item=None):
         print(f"prepare {item_type_projects} {src_item.get(name_key)} for migrating to {self.__dst_space_id}")
-        src_item.pop(versioning_strategy_key, None)
         src_item.pop(deployment_process_id_key, None)
         src_item.pop(variable_set_id_key, None)
+        # TODO once the migration supports clone packages; we do not have to pre-process VersioningStrategy
+        src_item.get(versioning_strategy_key)[donor_package_key] = None
+        src_item.get(versioning_strategy_key)[donor_package_step_id_key] = None
 
     # we do not want to clone the child items first;
     # child items will be created automatically when parent item is created
@@ -89,14 +102,29 @@ class Migration:
               f"{self.__dst_space_id}")
         src_item.pop(variable_set_id_key, None)
 
-    # SecretKey: NewValue: null
-    # must be replaced by a string
+    # SecretKey: NewValue: null; must be replaced by a placeholder string
     def __prepare_feed(self, src_item=None):
         print(f"prepare {item_type_feeds} {src_item.get(name_key)} for migrating to {self.__dst_space_id}")
         if secret_key_key in src_item and not src_item.get(secret_key_key).get(new_value_key):
             src_item.get(secret_key_key)[new_value_key] = hyphen_sign
             print(f"assigned a placeholder {hyphen_sign} to {new_value_key} for {secret_key_key}")
 
+    # Token: NewValue: null; must be replaced by a placeholder string
+    def __prepare_account(self, src_item=None):
+        print(f"prepare {item_type_accounts} {src_item.get(name_key)} for migrating to {self.__dst_space_id}")
+        if token_key in src_item and not src_item.get(token_key).get(new_value_key):
+            src_item.get(token_key)[new_value_key] = hyphen_sign
+            print(f"assigned a placeholder {hyphen_sign} to {new_value_key} for {token_key}")
+
+    # runbook is a new type inroduced in Octopus 2019.11; the older Octopus server may not support it
+    # TODO "RunbookProcessId": "RunbookProcess-Runbooks-68" the process needs to be put as post_process
+    # Need to upgrade Octopus server to the latest version
+    def __prepare_runbook(self, src_item=None):
+        print(f"prepare {item_type_runbooks} {src_item.get(name_key)} {src_item.get(id_key)} for migrating to "
+              f"{self.__dst_space_id}: popping {runbook_process_id_key}")
+        src_item.pop(runbook_process_id_key, None)
+        src_item.pop(published_runbook_snapshot_id_key, None)
+
     def __clone_item_to_space(self, item_type=None, item_name=None, item_id=None):
         item_badge = item_name if item_name else item_id
         print(f"===== clone {item_type} {item_badge} from {self.__src_space_id} to {self.__dst_space_id}")
@@ -137,12 +165,19 @@ class Migration:
         self.__replace_ids(dict_list=src_item_copy)
 
         print(f"check if {item_type} {item_badge} already exists in {self.__dst_space_id}")
-        dst_item = self.__find_matched_dst_item_by_src_item(item=src_item_copy, item_type=item_type)
+        dst_item = self.__find_matched_dst_item_by_src_item(src_item_dst_ids=src_item_copy, item_type=item_type)
         dst_item_exist = True if dst_item else False
         if dst_item_exist:
             if config.overwrite:
                 print(f"{self.__dst_space_id} already has {item_type} {item_badge}, overwriting it...")
                 src_item_copy[id_key] = dst_item.get(id_key)
+
+                # TODO bug in Octopus: PUT a runbook with null RunbookProcessId will remove RunbookProcessId from dst
+                if item_type == item_type_runbooks and src_item.get(runbook_process_id_key):
+                    matched_runbook_process_id = runbook_process_prefix + hyphen_sign + dst_item.get(id_key)
+                    print(f"Reassign {matched_runbook_process_id} for {src_item_copy.get(id_key)}")
+                    src_item_copy[runbook_process_id_key] = matched_runbook_process_id
+
                 dst_item = put_single_item_save(item_type=item_type, payload=src_item_copy,
                                                 space_id=self.__dst_space_id)
             else:
@@ -163,7 +198,7 @@ class Migration:
 
         return dst_id_value
 
-    def __create_a_type_to_another_space(self, item_type=None):
+    def __clone_type_to_space(self, item_type=None):
         if not item_type:
             raise ValueError("item_type must not be empty!")
         print(f"===== creating {item_type} from {self.__src_space_id} to {self.__dst_space_id}")
@@ -180,6 +215,20 @@ class Migration:
         dst_list_items = get_list_items_from_all_items(all_items=dst_all_items)
         self.__type_dst_list_items_dict[item_type] = dst_list_items
 
+    # some id like "Environments-123" should be removed if this is not referenced anyway
+    # most likely the entity has been removed and the reference is broken
+    # TODO it is a dangerous operation; find a better way to warn people to fix the source space first
+    def __check_broken_item_id(self, string=None):
+        # if the string is one of the src or dst item ids, it is valid
+        if string in self.__src_id_vs_dst_id_dict.keys() or string in self.__src_id_vs_dst_id_dict.values():
+            return False
+        for id_prefix, item_type in item_id_prefix_to_type_dict.items():
+            pattern = r"{}".format("^" + id_prefix + positive_integer_regex)
+            if re.match(pattern, string):
+                print(f"Please fix your source space by removing the broken id {string}")
+                return True
+        return False
+
     # recursively replace the old link id with the new link id
     # issue: if someone crazy names an environment name as "Environments-1", not "Development" or "Prod" etc,
     # it could cause the environment renamed as "Environment-10"
@@ -197,6 +246,10 @@ class Migration:
                     elif value in self.__src_id_payload_dict:
                         dict_list[key] = self.__clone_item_to_space(item_type=self.__src_id_type_dict.get(value),
                                                                     item_id=value)
+                    # remove the broken reference ids
+                    elif self.__check_broken_item_id(string=value):
+                        print(f"{value} does not exist, so assign null value to it")
+                        dict_list[key] = None
                 else:
                     self.__replace_ids(dict_list=value)
                 if isinstance(key, str):
@@ -206,21 +259,32 @@ class Migration:
                     elif key in self.__src_id_payload_dict:
                         new_key = self.__clone_item_to_space(item_type=self.__src_id_type_dict.get(key), item_id=key)
                         dict_list[new_key] = dict_list.pop(key)
+                    # remove the broken reference ids
+                    elif self.__check_broken_item_id(string=key):
+                        print(f"Key {value} does not exist, so pop it")
+                        dict_list.pop(key)
                 else:
-                    # TODO add a log if key is a dict; this case is very special
+                    print(f"{key} is a dict and needs further processing")
                     self.__replace_ids(dict_list=key)
         elif isinstance(dict_list, list):
-            for index, element in enumerate(dict_list):
+            print(f"dict_list is a list now, replace ids or delete broken ids")
+            # we must do reversely to avoid unexpected result on deleting by index
+            for index in range(len(dict_list) - 1, -1, -1):
+                element = dict_list[index]
                 if isinstance(element, str):
                     if element in self.__src_id_vs_dst_id_dict:
                         dict_list[index] = self.__src_id_vs_dst_id_dict.get(element)
                     elif element in self.__src_id_payload_dict:
                         dict_list[index] = self.__clone_item_to_space(item_type=self.__src_id_type_dict.get(element),
                                                                       item_id=element)
+                    elif self.__check_broken_item_id(string=element):
+                        print(f"delete a broken reference id: {element}; index {index}")
+                        del dict_list[index]
                 else:
                     self.__replace_ids(dict_list=element)
+        # # TODO log type: None, boolean, integer, float etc
         # else:
-        #     # None, boolean, integer, float etc
+        #     print(f"the type of the object is {type(dict_list)}")
 
     def __clone_child(self, src_parent_id=None, dst_parent_id=None, child_type=None, child_id_key=None):
         # source item
@@ -239,7 +303,7 @@ class Migration:
         src_child = self.__src_id_payload_dict.get(src_child_id)
         # TODO Octopus bug https://help.octopus.com/t/504-gateway-time-out-on-getting-all-variables/24732
         if not src_child:
-            print(f"{src_child_id} does not exist in the local all_*.yaml file, so get it from {self.__src_space_id}")
+            print(f"{src_child_id} does not exist in the memory, so get it from {self.__src_space_id}")
             src_child = get_single_item_by_name_or_id_save(item_type=child_type, item_id=src_child_id,
                                                            space_id=self.__src_space_id)
             self.__src_id_payload_dict[src_child_id] = src_child
@@ -275,12 +339,20 @@ class Migration:
         self.__clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_variables,
                            child_id_key=variable_set_id_key)
 
+    def __post_process_runbook(self, src_id=None, dst_id=None):
+        # TODO buggy post/put runbooks; some of the
+
+        print(f"clone {item_type_runbook_processes} from {item_type_runbooks} {src_id} to {dst_id} in "
+              f"{self.__dst_space_id}")
+        self.__clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_runbook_processes,
+                           child_id_key=runbook_process_id_key)
+
     # tenant variables is special and its id is also tenant id, such as "Tenants-401"
     # so you have to use a separate map to store the tenants variables
     # also put/post tenant variables uses a different url from put variables
     def __post_process_tenant_variables(self, src_id=None, dst_id=None):
-        print(f"clone {item_type_tenant_variables} from {item_type_tenants} {src_id} to {dst_id} in "
-              f"{self.__dst_space_id}")
+        print(f"clone {item_type_tenant_variables} from {item_type_tenants} {src_id} in {self.__src_space_id} "
+              f"to {dst_id} in {self.__dst_space_id}")
         src_tenant_variables = self.__src_tenant_variables_payload_dict.get(src_id)
         src_tenant_variables_copy = copy.deepcopy(src_tenant_variables)
 
@@ -299,7 +371,8 @@ class Migration:
         src_tag = self.__src_id_payload_dict.get(item_id)
         src_tag_copy = copy.deepcopy(src_tag)
         src_tag_copy.pop(id_key, None)
-        dst_tag_set = self.__find_matched_dst_item_by_src_item(item=src_tag_set, item_type=item_type_tag_sets)
+        dst_tag_set = self.__find_matched_dst_item_by_src_item(src_item_dst_ids=src_tag_set,
+                                                               item_type=item_type_tag_sets)
         if dst_tag_set:
             print(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} exists in {self.__dst_space_id}, "
                   f"so try to find {item_id}")
@@ -311,8 +384,8 @@ class Migration:
                                                    space_id=self.__dst_space_id,
                                                    overwrite=True)
             else:
-                print("{item_type_tag_sets} {dst_tag_set.get(name_key)} already has {item_id} in {self.__dst_space_id},"
-                      " skip")
+                print(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} already has {item_id} in "
+                      f"{self.__dst_space_id}, skip")
         else:
             print(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} does not exist in {self.__dst_space_id}, "
                   f"so create it with {item_type_tags} {item_id}")
@@ -326,17 +399,27 @@ class Migration:
                                                         value=item_id)
         return item_id
 
-    def __get_all_local_types(self, item_types=item_types_inside_space):
-        print(f"getting {item_types} from space {self.__src_space_id}...")
-        for item_type in item_types:
-            print(f"Retrieving {item_type} from space {self.__src_space_id}...")
-            src_list_items = get_one_type_to_list_ignore_error(item_type=item_type, space_id=self.__src_space_id)
-            if not src_list_items:
-                # for cloning space from another Octopus server
-                print(f"{item_type} cannot be downloaded from space {self.__src_space_id}, so read local file...")
+    def __load_types(self, fake_space=False):
+        if fake_space:
+            print(f"Reading files {self.__item_types_set} from fake space {self.__src_space_id}...")
+        else:
+            print(f"Downloading {self.__item_types_set} from space {self.__src_space_id}...")
+        actual_src_space_id = None
+        for item_type in self.__item_types_set:
+            # for cloning space from another Octopus server
+            if fake_space:
+                print(f"Loading {item_type} in source fake space {self.__src_space_id} from the local file...")
                 src_list_items = get_list_items_from_file(item_type=item_type, space_id=self.__src_space_id)
+            # for cloning space to space on the same Octopus server
+            else:
+                print(f"Loading {item_type} from source space {self.__src_space_id}...")
+                src_list_items = get_one_type_to_list(item_type=item_type, space_id=self.__src_space_id)
             self.__type_src_list_items_dict[item_type] = src_list_items
             for src_item in src_list_items:
+                if not actual_src_space_id and fake_space and item_type == item_type_projects \
+                        and src_item.get(space_id_key):
+                    actual_src_space_id = src_item.get(space_id_key)
+                    print(f"the actual source space id is {actual_src_space_id}; the fake one is {self.__src_space_id}")
                 if src_item.get(id_key):
                     self.__src_id_payload_dict[src_item.get(id_key)] = src_item
                     self.__src_id_type_dict[src_item.get(id_key)] = item_type
@@ -345,8 +428,11 @@ class Migration:
                 else:
                     pprint(src_item)
                     raise ValueError(f"{item_type} does not have valid id")
+        if fake_space and not actual_src_space_id:
+            raise ValueError(f"Could not find an actual space id inside the fake space {self.__src_space_id}")
+        return actual_src_space_id if actual_src_space_id else self.__src_space_id
 
-    def __prep_all_tag_sets(self):
+    def __prep_tag_sets(self):
         print("prepare for all tags")
         src_list_tag_sets = self.__type_src_list_items_dict.get(item_type_tag_sets)
         for src_tag_set in src_list_tag_sets:
@@ -355,25 +441,32 @@ class Migration:
                 self.__src_id_payload_dict[canonical_tag_name] = src_tag
                 self.__src_id_type_dict[canonical_tag_name] = item_type_tags
 
-    def __initialize_maps(self, src_space_id=None, dst_space_id=None, item_types=item_types_inside_space):
+    def __initialize_maps(self, src_space_id=None, dst_space_id=None, fake_space=False):
         self.__src_space_id = src_space_id
         self.__dst_space_id = dst_space_id
-        self.__src_id_vs_dst_id_dict[src_space_id] = dst_space_id
 
         self.__type_prep_func_dict[item_type_tag_sets] = self.__prepare_tag_set
         self.__type_prep_func_dict[item_type_projects] = self.__prepare_project
         self.__type_prep_func_dict[item_type_library_variable_sets] = self.__prepare_library_variable_set
         self.__type_prep_func_dict[item_type_feeds] = self.__prepare_feed
+        self.__type_prep_func_dict[item_type_runbooks] = self.__prepare_runbook
+        self.__type_prep_func_dict[item_type_accounts] = self.__prepare_account
 
         self.__type_post_func_dict[item_type_projects] = self.__post_process_project
         self.__type_post_func_dict[item_type_library_variable_sets] = self.__post_process_library_variable_set
         self.__type_post_func_dict[item_type_tenants] = self.__post_process_tenant_variables
+        self.__type_post_func_dict[item_type_runbooks] = self.__post_process_runbook
 
         self.__type_full_func_dict[item_type_tags] = self.__full_process_tags
 
-        self.__get_all_local_types(item_types=item_types)
+        actual_src_space_id = self.__load_types(fake_space=fake_space)
 
-        self.__prep_all_tag_sets()
+        if fake_space:
+            self.__src_id_vs_dst_id_dict[actual_src_space_id] = dst_space_id
+        else:
+            self.__src_id_vs_dst_id_dict[src_space_id] = dst_space_id
+
+        self.__prep_tag_sets()
 
     def __save_space_map(self):
         current_time = strftime("%Y-%m-%d-%H-%M-%S", gmtime())
@@ -381,9 +474,9 @@ class Migration:
                                                 item_type=item_type_migration, space_id=self.__dst_space_id)
         save_file(file_path_name=local_file, content=self.__src_id_vs_dst_id_dict)
 
-    def clone_space(self, src_space_id=None, dst_space_id=None, item_types_comma_delimited=None):
+    def clone_space(self, src_space_id=None, dst_space_id=None, item_types_comma_delimited=None, fake_space=False):
         if item_types_comma_delimited:
-            process_types = must_have_types + item_types_comma_delimited.split(",")
+            process_types = must_have_types + item_types_comma_delimited.split(comma_sign)
         else:
             process_types = normal_cloneable_types
         if not config.overwrite:
@@ -391,30 +484,29 @@ class Migration:
                                      f"Some entities may already exist in {dst_space_id}; "
                                      f"Do you want to overwrite the existing entities? "
                                      f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'
-        item_types_set = set(normal_cloneable_types + process_types +
-                             [item_type_deployment_processes, item_type_variables, item_type_tenant_variables])
-        self.__initialize_maps(src_space_id=src_space_id, dst_space_id=dst_space_id, item_types=item_types_set)
+        self.__item_types_set = set(normal_cloneable_types + process_types +
+                                    [item_type_deployment_processes, item_type_variables, item_type_tenant_variables])
+        self.__initialize_maps(src_space_id=src_space_id, dst_space_id=dst_space_id, fake_space=fake_space)
         print(f"creating types {process_types} from {src_space_id} to {dst_space_id}")
         for item_type in process_types:
             if item_type in normal_cloneable_types:
-                self.__create_a_type_to_another_space(item_type=item_type)
+                self.__clone_type_to_space(item_type=item_type)
         self.__save_space_map()
 
-    def clone_space_item(self, src_space_id=None, dst_space_id=None, item_type=None, item_name=None, item_id=None):
+    def clone_space_item(self, src_space_id=None, dst_space_id=None, item_type=None, item_name=None, item_id=None,
+                         fake_space=False):
         item_badge = item_name if item_name else item_id
-        if not item_badge:
-            raise ValueError(f"either item_name or item_id is required")
         if not config.overwrite:
             config.overwrite = input(
                 f"***** You are cloning {item_type} {item_badge} from {src_space_id} to {dst_space_id}; "
                 f"Some entities may already exist in {dst_space_id}; "
                 f"Do you want to overwrite the existing entities? "
                 f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'
-        item_types_set = set(normal_cloneable_types + [item_type, item_type_deployment_processes, item_type_variables,
-                                                       item_type_tenant_variables])
-        self.__initialize_maps(src_space_id=src_space_id, dst_space_id=dst_space_id, item_types=item_types_set)
+        self.__item_types_set = set(normal_cloneable_types + [item_type, item_type_deployment_processes,
+                                                              item_type_variables, item_type_tenant_variables])
+        self.__initialize_maps(src_space_id=src_space_id, dst_space_id=dst_space_id, fake_space=fake_space)
         for must_have_type in must_have_types:
-            self.__create_a_type_to_another_space(item_type=must_have_type)
+            self.__clone_type_to_space(item_type=must_have_type)
         if item_type in normal_cloneable_types:
             self.__clone_item_to_space(item_type=item_type, item_name=item_name, item_id=item_id)
         self.__save_space_map()
